function varargout=bench(Lfun,setfun,In,varargin)
% FUNCTION fc_bench.bench
% -- bench(LFUN,SETFUN,IN)
%     Runs benchmark for each function given in the cell array LFUN. The 
%     function handle SETFUN is used to set input datas to these functions. 
%     There is the imposed syntax:
%        function [INPUTS,BDS,ERRORS]=SETFUN(N,verbose,varargin)
%          ...
%        end
%     By default, computational time in second of 
%     each function in LFUN with INPUTS given by
%        [INPUTS,BDS]=SETFUN(N,verbose,varargin{:})  for all N in IN
%     is evaluated by tic-toc command:
%        t=tic(); OUT=LFUN{i}( INPUTS{:} ); tcpu=toc(t);
%     
%     Example to benchmark the function MTIMES (matricial product) between two 
%       N-by-N matrices for N in 50:50:200: 
%         LFUN={ @(X,Y)mtimes(X,Y) };  % or LFUN={ @mtimes };
%         SETFUN=@(varargin) set_matprod(varargin{:}); % or SETFUN=@set_matprod
%         IN=50:50:200;
%         fc_bench.bench(LFUN, SETFUN, IN); 
%     
%       The 'set_matprod' function  as second parameter is given by
%         function [Inputs,bDs,Errors]=set_matprod(N,verbose,varargin)
%           X=randn(N,N); Y=randn(N,N);
%           bDs(1)=fc_bench.bdata('N',N,'%d',5);
%           Inputs={X,Y};
%           Errors={};
%         end

% -- bench(Lfun,setfun,In, key,value,...)
%
%     See User's Guide
%
  p = inputParser;
  p.KeepUnmatched=true; 
  p.PartialMatching=false;
  p.addParamValue('nbruns',5);
  p.addParamValue('debug',false,@islogical);
  p.addParamValue('compfun',{}); % Use only if all
  p.addParamValue('info',true,@islogical);
  p.addParamValue('labelsinfo',false,@islogical);
  p.addParamValue('save',false,@islogical);
  p.addParamValue('before',{},@iscell); % empty or length(before)==length(Lfun)
  p.addParamValue('after',{},@iscell);  % empty or length(before)==length(Lfun)
  p.addParamValue('comment','',@(x) ischar(x) || iscell(x));
  p.addParamValue('savefile','',@ischar);
  p.addParamValue('autotag',true,@islogical);
  p.addParamValue('savedir','benchs',@ischar); % relative to currrent path. default 'benchs'
  p.addParamValue('names',{},@iscell); % same length as Lfun : to specify name of each function
  p.parse(varargin{:});
  R=p.Results;
  varargin=fc_tools.utils.deleteCellOptions(varargin,p.Parameters);
  DateNum=now;
  
  Debug=R.debug;nbruns=R.nbruns;
  compfun=R.compfun;
  if fc_tools.utils.isfunhandle(compfun), compfun={compfun};end
  
  savefile=getFullFileName(R,Lfun);
  if ~isempty(savefile)
    [fid,res]=fopen(savefile,'w');
    assert(fid>=0,res)
    Fprintf=@(varargin) EchoingFprintf(fid,varargin{:});
  else
    Fprintf=@(varargin) fprintf(varargin{:});
  end
  
  ltd=fc_tools.utils.line_text_delimiter();
  format='';
  Fprintf(ltd);
  if R.info
    %Fprintf(ltd);
    fc_bench.print_info('fprintf',Fprintf);
    Fprintf(ltd);
  end
  
  [In,getIn,nIn]=setIn(In);
  [datas,setbDs,Errors]=setfun(getIn(In,1),true,'fprintf',Fprintf,varargin{:}); % 
  Fprintf(ltd);
  
  names=set_names(R.names,Lfun);
  
  % Print comment
  if length(R.comment)>0
    PrintCellsOfString(R.comment,Fprintf)
    Fprintf(ltd);
  end
  
  if ~isempty(R.compfun)
    if fc_tools.utils.isfunhandle(R.compfun), R.compfun={R.compfun};end
  end
  if R.labelsinfo
    Print_labelsinfo(names,Lfun,R.compfun,Fprintf)
    Fprintf(ltd);
  end
  
  if R.save
    Fprintf('#benchfile: %s\n',savefile);
  end
   
  Before=R.before;
  After=R.after;
  
  Res=setBenchColumns(setbDs,Lfun,names,Before,After,Errors,compfun,nIn);
  [format,nlabels]=fc_bench.print_title(Res,DateNum,nbruns,'fprintf',Fprintf);
  for row=1:nIn
    N=getIn(In,row);
    Res=computeBenchColumns(Res,@() setfun(N,false,varargin{:}),Lfun,names,Before,After,compfun,nbruns,row);
    PrintRow(Res,Fprintf,nlabels,row);
  end
  if ~isempty(savefile)
    fclose(fid);
  end
  if nargout==1
    varargout{1}=Res;
  end
end

function EchoingFprintf(fid,varargin)
  fprintf(varargin{:});
  fprintf(fid,varargin{:});
end

function str=string_command(Lfun,setfun,In,varargin)
  str='fc_bench.bench( { ';
  for i=1:length(Lfun)-1
    str=[str,fc_tools.utils.fun2str(Lfun{i}),', '];
  end
  str=[str,fc_tools.utils.fun2str(Lfun{end}),' },  '];
  str=[str,fc_tools.utils.fun2str(setfun),',In, ...'];
  str=[str,')'];
end

function [In,getIn,nIn]=setIn(In)
  if isnumeric(In)
    if size(In,2)==1,In=In.';end
  end
  if size(In,1)==1
    if iscell(In) 
      getIn=@(In,i) In{i};
    else
      getIn=@(In,i) In(i);
    end
    nIn=size(In,2);
  else
    if iscell(In), getIn=@(In,i) {In{i,:}};else;  getIn=@(In,i) In(i,:);end
    nIn=size(In,1);
  end
end

function savefile=getFullFileName(R,Lfun)
  savefile='';
  if R.save || ~isempty(R.savefile) 
    savedir=R.savedir;
    if isempty(savedir), savedir='.';end
    savefile=R.savefile;
    if isempty(R.savefile)
      name=fc_bench.funname_small(Lfun{1});
      I=strfind(name,'_');
      if length(I)>0, name=name(1:I(end)-1);end
      Date=datestr(now,'yyyy-mm-dd_HH-MM-SS');
      [success,message,messageid]=mkdir('.',savedir);
      if ~success, error(message);end
      if R.autotag
        [s,r]=fc_tools.sys.getSoftware();r=strrep(r,'.','');
        savefile=fullfile(savedir,['bench_',name,'_',s,r,'_',Date,'.out']);
      else
        savefile=fullfile(savedir,['bench_',name,'_',Date,'.out']);
      end
    else
      [pathstr,filename,ext] = fileparts(R.savefile);
      if isempty(pathstr)
        [success,message,messageid]=mkdir('.',savedir);
        if ~success, error(message);end
        savefile=fullfile(savedir,R.savefile);
      else
        [success,message,messageid]=mkdir('.',pathstr);
        if ~success, error(message);end
        savefile=R.savefile;
      end
    end
  end
end

function n=LengthCell(B,nfun)
  n=0;
  assert(iscell(B))
  if isempty(B), return;end
  assert(length(B)==nfun)
  for i=1:nfun
    n=n+length(B{i}); % B{i} is a function handle or cell array of function handle
  end
end

function PrintCellsOfString(C,Fprintf)
  if ischar(C)
    Fprintf(C);
    if ~strcmp(C(end-1:end),'\n'), Fprintf('\n');end
  else % iscell
    for i=1:length(C)
      Fprintf(C{i});
      if ~strcmp(C{i}(end-1:end),'\n'), Fprintf('\n');end
    end
  end
end

function Print_labelsinfo(names,Lfun,compfun,Fprintf)
  nfun=length(Lfun);
  Fprintf('# Benchmarking functions:\n')
  for i=1:nfun
    Fprintf('#   fun[%d], %8s: %s\n',i-1,names{i},func2str(Lfun{i}))
  end
  if ~isempty(compfun)
    assert(iscell(compfun))
    ltd=fc_tools.utils.line_text_delimiter();
    Fprintf(ltd);
    Fprintf('# Comparative functions:\n')
    if length(compfun)==1
      Fprintf('#   comp[i-1], compares outputs of fun[0] and fun[i] by using \n#       %s\n',func2str(compfun{1}))
      Fprintf('#   where  \n')  
      Fprintf('#      - 1st input parameter is the output of fun[0]\n')
      Fprintf('#      - 2nd input parameter is the output of fun[i]\n')
    else
      for s=1:length(compfun)
        Fprintf('#   comp[i-1,%d], compares outputs of fun[0] and fun[i]\n#       %s\n',s-1,func2str(compfun{s}))
      end
      Fprintf('#    For each comparative function:\n')
      Fprintf('#      - 1st input parameter is the output of fun[0]\n')
      Fprintf('#      - 2nd input parameter is the output of fun[i]\n')
    end
  end
end

function PrintRow(Res,Fprintf,nlabels,row)
  line='';
  for k=1:length(Res)
    vs=strtrim(sprintf(Res{k}.sformat,Res{k}.values{row}));
    vs=[repmat(' ',1,nlabels(k)-length(line)-length(vs)),vs];
    line=[line,vs];
  end
  Fprintf('%s\n',line);
end

function names=set_names(names,Lfun)
  nfun=length(Lfun);
  if isempty(names), names=cell(1,nfun);end
  assert(length(names)==nfun)  
  for i=1:nfun
    if isempty(names{i})
      names{i}=fc_bench.funname_small(Lfun{i}); 
    end
  end
end

function Res=setBenchColumns(setbDs,Lfun,names,Before,After,Error,compfun,nIn)
  nfun=length(Lfun);
  k=1;
  for i=1:length(setbDs)
    Res{k}=fc_bench.bdatacol_bdata(setbDs{i},nIn);
    k=k+1;
  end
  
  for i=1:nfun
    if ~isempty(Before)
      for j=1:length(Before{i})
        Res{k}=fc_bench.bdatacol_bdata(Before{i}{j},nIn);
        k=k+1;
      end
    end
    Res{k}=fc_bench.bdatacol('name',[names{i},'(s)'],'values',cell(1,nIn),'sformat','%.3f','strlen',length(names{i})+6);
    k=k+1;
    if ~isempty(After)
      for j=1:length(After{i})
        Res{k}=fc_bench.bdatacol_bdata(After{i}{j},nIn);
        k=k+1;
      end
    end
    if ~isempty(Error)
      nE=length(Error);
      if nE>1, fname=@(i,j) sprintf('  Error[%d,%d]',i-1,j-1);else fname=@(i,j) sprintf('  Error[%d]',i-1);end
      for j=1:nE
        name=fname(i,j);
        Res{k}=fc_bench.bdatacol('name',name,'values',cell(1,nIn),'sformat','%.3e','strlen',length(name)+3,'fun',Error{j});
        k=k+1;
      end
    end
    if i>1
      if ~isempty(compfun)
        nE=length(compfun);
        if nE>1, fname=@(i,j) sprintf('  comp[%d,%d]',i-2,j-1);else fname=@(i,j) sprintf('  comp[%d]',i-2);end
        for j=1:nE
          name=fname(i,j);
          Res{k}=fc_bench.bdatacol('name',name,'values',cell(1,nIn),'sformat','%.3e','strlen',length(name)+3,'fun',compfun{j});
          k=k+1;
        end
      end
    end
  end
  
end

function Res=computeBenchColumns(Res,setfun,Lfun,names,Before,After,compfun,nbruns,row)
  [datas,bDs,Error]=setfun();
  nfun=length(Lfun);
  k=1;
  for i=1:nfun
    [tcpu,Out]=fc_bench.mean_run(Lfun{i},datas,nbruns);
    if i==1 
      OutRef=Out; 
      for j=1:length(bDs)
        Res{k}.values{row}=bDs{j}.get_value(Out);
%          if fc_tools.utils.isfunhandle(bDs{j}.value)
%            Res{k}.values{row}=bDs{j}.value(Out);
%          else
%            Res{k}.values{row}=bDs{j}.value;
%          end
        k=k+1;
      end
    end
    if ~isempty(Before)
      for j=1:length(Before{i})
        %Res{k}.values{row}=Res{k}.fun(Out);
        Res{k}.values{row}=Before{i}{j}.get_value(Out);
%          if fc_tools.utils.isfunhandle(bDs{j}.value)
%            if iscell(Out), Res{k}.values{row}=Before{i}{j}.value(Out{:});else, Res{k}.values{row}=Before{i}{j}.value(Out);end
%          else
%            if iscell(Out), Res{k}.values{row}=Before{i}{j}(Out{:});else, Res{k}.values{row}=Before{i}{j}(Out);end
%          end
        k=k+1;
      end
    end
    Res{k}.values{row}=tcpu;
    k=k+1;

    if ~isempty(After)
      for j=1:length(After{i})
        Res{k}.values{row}=After{i}{j}.get_value(Out);
        %if iscell(Out), Res{k}.values{row}=After{i}{j}(Out{:});else, Res{k}.values{row}=After{i}{j}(Out);end
        k=k+1;
      end
    end
    if ~isempty(Error)
      nE=length(Error);
      for j=1:nE
        if iscell(Out), Res{k}.values{row}=Error{j}(Out{:});else, Res{k}.values{row}=Error{j}(Out);end
        k=k+1;
      end
    end
    if i>1
      if ~isempty(compfun)
        nE=length(compfun);
        for j=1:nE
          Res{k}.values{row}=compfun{j}(OutRef,Out);
          k=k+1;
        end
      end
    end
  end
end
